@@ -9,7 +9,7 @@
 #
 package Redis;
 {
-  $Redis::VERSION = '1.972';
+  $Redis::VERSION = '1.974';
 }
 
 # ABSTRACT: Perl binding for Redis database
@@ -38,6 +38,7 @@ use constant WIN32       => $^O =~ /mswin32/i;
 use constant EWOULDBLOCK => eval {Errno::EWOULDBLOCK} || -1E9;
 use constant EAGAIN      => eval {Errno::EAGAIN} || -1E9;
 use constant EINTR       => eval {Errno::EINTR} || -1E9;
+use constant BUFSIZE     => 4096;
 
 sub _maybe_enable_timeouts {
     my ($self, $socket) = @_;
@@ -54,8 +55,9 @@ sub _maybe_enable_timeouts {
 
 sub new {
   my ($class, %args) = @_;
-  my $self  = bless {}, $class;
+  my $self = bless {}, $class;
 
+  $self->{__buf} = '';
   $self->{debug} = $args{debug} || $ENV{REDIS_DEBUG};
 
   ## Deal with REDIS_SERVER ENV
@@ -217,7 +219,19 @@ sub __std_cmd {
   # If this is an EXEC command, in pipelined mode, and one of the commands
   # executed in the transaction yields an error, we must collect all errors
   # from that command, rather than throwing an exception immediately.
-  my $collect_errors = $cb && uc($command) eq 'EXEC';
+  my $uc_command = uc($command);
+  my $collect_errors = $cb && $uc_command eq 'EXEC';
+
+  if ($uc_command eq 'MULTI') {
+      $self->{__inside_transaction} = 1;
+  } elsif ($uc_command eq 'EXEC' || $uc_command eq 'DISCARD') {
+      delete $self->{__inside_transaction};
+      delete $self->{__inside_watch};
+  } elsif ($uc_command eq 'WATCH') {
+      $self->{__inside_watch} = 1;
+  } elsif ($uc_command eq 'UNWATCH') {
+      delete $self->{__inside_watch};
+  }
 
   ## Fast path, no reconnect;
   $self->{reconnect}
@@ -244,6 +258,9 @@ sub __with_reconnect {
       ref($_) eq 'Redis::X::Reconnect'
         or die $_;
 
+      $self->{__inside_transaction} || $self->{__inside_watch}
+        and croak("reconnect disabled inside transaction or watch");
+
       $self->connect;
       $cb->();
     }
@@ -312,7 +329,7 @@ sub quit {
     $self->__send_command('QUIT');
   };
 
-  close(delete $self->{sock}) if $self->{sock};
+  $self->__close_sock() if $self->{sock};
 
   return 1;
 }
@@ -328,7 +345,7 @@ sub shutdown {
 
   $self->wait_all_responses;
   $self->__send_command('SHUTDOWN');
-  close(delete $self->{sock}) || croak("Can't close socket: $!");
+  $self->__close_sock() || croak("Can't close socket: $!");
 
   return 1;
 }
@@ -347,7 +364,7 @@ sub ping {
     $self->__std_cmd('PING');
   }
   catch {
-    close(delete $self->{sock});
+    $self->__close_sock();
     return;
   };
 }
@@ -424,22 +441,21 @@ sub wait_for_messages {
       $s->add($sock);
 
       while ($s->can_read($timeout)) {      
-        my $has_stuff = __try_read_sock($sock);
+        my $has_stuff = $self->__try_read_sock($sock);
         # If the socket is ready to read but there is nothing to read, ( so
         # it's an EOF ), try to reconnect.
         defined $has_stuff
           or $self->__throw_reconnect('EOF from server');
-        while (1) {
-          my $has_stuff = __try_read_sock($sock);
-          $has_stuff
-            or last ; ## no data ( or socket became EOF), back to select until
-                      ## timeout
-      
+
+        do {
           my ($reply, $error) = $self->__read_response('WAIT_FOR_MESSAGES');
           croak "[WAIT_FOR_MESSAGES] $error, " if defined $error;
           $self->__process_pubsub_msg($reply);
           $count++;
-        }
+
+          # if __try_read_sock() return 0 (no data)
+          # or undef ( socket became EOF), back to select until timeout
+        } while ($self->{__buf} || $self->__try_read_sock($sock));
       }
     
     });
@@ -565,6 +581,8 @@ sub __is_valid_command {
 sub connect {
   my ($self) = @_;
   delete $self->{sock};
+  delete $self->{__inside_watch};
+  delete $self->{__inside_transaction};
 
   # Suppose we have at least one command response pending, but we're about
   # to reconnect.  The new connection will never get a response to any of
@@ -597,6 +615,8 @@ sub __build_sock {
   $self->{sock} = $self->{builder}->($self)
     || croak("Could not connect to Redis server at $self->{server}: $!");
 
+  $self->{__buf} = '';
+
   if (exists $self->{password}) {
     try { $self->auth($self->{password}) }
     catch {
@@ -610,6 +630,14 @@ sub __build_sock {
   return;
 }
 
+sub __close_sock {
+  my ($self) = @_;
+  $self->{__buf} = '';
+  delete $self->{__inside_watch};
+  delete $self->{__inside_transaction};
+  return close(delete $self->{sock});
+}
+
 sub __on_connection {
 
     my ($self) = @_;
@@ -676,7 +704,7 @@ sub __send_command {
   }
 
   ## Check to see if socket was closed: reconnect on EOF
-  my $status = __try_read_sock($sock);
+  my $status = $self->__try_read_sock($sock);
   $self->__throw_reconnect('Not connected to any server')
     unless defined $status;
 
@@ -747,7 +775,7 @@ sub __read_line {
   my $self = $_[0];
   my $sock = $self->{sock};
 
-  my $data = <$sock>;
+  my $data = $self->__read_line_raw;
   croak("Error while reading from Redis server: $!")
     unless defined $data;
 
@@ -758,122 +786,100 @@ sub __read_line {
   return ($type, $data);
 }
 
-sub __read_len {
-  my ($self, $len) = @_;
+sub __read_line_raw {
+  my $self = $_[0];
+  my $sock = $self->{sock};
+  my $buf = \$self->{__buf};
 
-  my $data   = '';
-  my $offset = 0;
-  while ($len) {
-    my $bytes = read $self->{sock}, $data, $len, $offset;
-    croak("Error while reading from Redis server: $!")
-      unless defined $bytes;
-    croak("Redis server closed connection") unless $bytes;
-
-    $offset += $bytes;
-    $len -= $bytes;
+  if (length $$buf) {
+    my $idx = index($$buf, "\r\n");
+    $idx >= 0 and return substr($$buf, 0, $idx + 2, '');
   }
 
+  while (1) {
+    my $bytes = sysread($sock, $$buf, BUFSIZE, length($$buf));
+    next if !defined $bytes && $! == EINTR;
+    return unless defined $bytes && $bytes;
+
+    # start looking for \r\n where we stopped last time
+    # extracting one is required to handle corner case
+    # where \r\n are split and therefore read by two conseqent sysreads
+    my $idx = index($$buf, "\r\n", length($$buf) - $bytes - 1);
+    $idx >= 0 and return substr($$buf, 0, $idx + 2, '');
+  }
+}
+
+sub __read_len {
+  my ($self, $len) = @_;
+  my $buf = \$self->{__buf};
+  my $buflen = length($$buf);
+
+  if ($buflen < $len) {
+    my $to_read = $len - $buflen;
+    while ($to_read > 0) {
+      my $bytes = sysread($self->{sock}, $$buf, BUFSIZE, length($$buf));
+      next if !defined $bytes && $! == EINTR;
+      croak("Error while reading from Redis server: $!") unless defined $bytes;
+      croak("Redis server closed connection") unless $bytes;
+      $to_read -= $bytes;
+    }
+  }
+
+  my $data = substr($$buf, 0, $len, '');
   chomp $data;
   warn "[RECV RAW] '$data'" if $self->{debug};
 
   return $data;
 }
 
-
-#
-# The reason for this code:
-#
-# IO::Select and buffered reads like <$sock> and read() don't mix
-# For example, if I receive two MESSAGE messages (from Redis PubSub),
-# the first read for the first message will probably empty to socket
-# buffer and move the data to the perl IO buffer.
-#
-# This means that IO::Select->can_read will return false (after all
-# the socket buffer is empty) but from the application point of view
-# there is still data to be read and process
-#
-# Hence this code. We try to do a non-blocking read() of 1 byte, and if
-# we succeed, we put it back and signal "yes, Virginia, there is still
-# stuff out there"
-#
-# We could just use sysread and leave the socket buffer with the second
-# message, and then use IO::Select as intended, and previous versions of
-# this code did that (check the git history for this file), but
-# performance suffers, about 20/30% slower, mostly because we do a lot
-# of "read one line", where <$sock> beats the crap of anything you can
-# write on Perl-land.
-#
 sub __try_read_sock {
-  my $sock = shift;
+  my ($self, $sock) = @_;
   my $data = '';
 
-  __fh_nonblocking($sock, 1);
+  while (1) {
+      # WIN32 doesn't support MSG_DONTWAIT,
+      # need to swith fh to nonblockng mode manually.
+      # For Unix still use MSG_DONTWAIT because of fewer syscalls
+      my ($res, $err);
+      if (WIN32) {
+          __fh_nonblocking_win32($sock, 1);
+          $res = recv($sock, $data, BUFSIZE, 0);
+          $err = 0 + $!;
+          __fh_nonblocking_win32($sock, 0);
+      } else {
+          $res = recv($sock, $data, BUFSIZE, MSG_DONTWAIT);
+          $err = 0 + $!;
+      }
 
-  ## Lots of problems with Windows here. This is a temporary fix until I
-  ## figure out what is happening there. It looks like the wrong fix
-  ## because we should not mix sysread (unbuffered I/O) with ungetc()
-  ## below (buffered I/O), so I do expect to revert this soon.
-  ## Call it a run through the CPAN Testers Gautlet fix. If I had to
-  ## guess (and until my Windows box has a new power supply I do have to
-  ## guess), I would say that the problems lies with the call
-  ## __fh_nonblocking(), where on Windows we don't end up with a non-
-  ## blocking socket.
-  ## See
-  ##  * https://github.com/melo/perl-redis/issues/20
-  ##  * https://github.com/melo/perl-redis/pull/21
-  my $len;
-  local $! = 0;
-  if (WIN32) {
-    $len = sysread($sock, $data, 1);
+      if (defined $res) {
+        ## have read some data
+        if (length($data)) {
+            $self->{__buf} .= $data;
+            return 1;
+        }
+
+        ## no data but also no error means EOF
+        return;
+      }
+
+      next if $err && $err == EINTR;
+
+      ## Keep going if nothing there, but socket is alive
+      return 0 if $err and ($err == EWOULDBLOCK or $err == EAGAIN);
+
+      ## result is undef but err is 0? should never happen
+      return if $err == 0;
+
+      ## For everything else, there is Mastercard...
+      croak("Unexpected error condition $err/$^O, please report this as a bug");
   }
-  else {
-    $len = read($sock, $data, 1);
-  }
-  my $err = 0 + $!;
-  __fh_nonblocking($sock, 0);
-
-  if (defined($len)) {
-    ## Have stuff
-    if ($len > 0) {
-      $sock->ungetc(ord($data));
-      return 1;
-    }
-    ## EOF according to the docs
-    ## (however FreeBSD seems to return $len = 0 when read fails)
-    elsif ($len == 0) {
-      if ($err and ($err == EWOULDBLOCK or $err == EAGAIN or $err == EINTR)) {
-        return 0;
-      } 
-      return;
-    }
-    else {
-      croak("read()/sysread() are really bonkers on $^O, return negative values ($len)");
-    }
-  }
-
-  ## Keep going if nothing there, but socket is alive
-  return 0 if $err and ($err == EWOULDBLOCK or $err == EAGAIN or $err == EINTR);
-
-  ## No errno, but result is undef?? This happens sometimes on my tests
-  ## when the server timesout the client. I traced the system calls and
-  ## I see the read() system call return 0 for EOF, but on this side of
-  ## perl, we get undef... We should see the 0 return code for EOF, I
-  ## suspect the fact that we are in non-blocking mode is the culprit
-  return if $err == 0;
-
-  ## For everything else, there is Mastercard...
-  croak("Unexpected error condition $err/$^O, please report this as a bug");
 }
 
-
-### Copied from AnyEvent::Util
-BEGIN {
-  *__fh_nonblocking = (WIN32)
-    ? sub($$) { ioctl $_[0], 0x8004667e, pack "L", $_[1]; }    # FIONBIO
-    : sub($$) { fcntl $_[0], F_SETFL, $_[1] ? O_NONBLOCK : 0; };
+## Copied from AnyEvent::Util
+sub __fh_nonblocking_win32 {
+    ioctl $_[0], 0x8004667e, pack "L", $_[1];
 }
 
-
 ##########################
 # I take exception to that
 
@@ -898,7 +904,7 @@ Redis - Perl binding for Redis database
 
 =head1 VERSION
 
-version 1.972
+version 1.974
 
 =head1 SYNOPSIS
 
@@ -924,10 +930,10 @@ version 1.972
     ## Enable auto-reconnect
     ## Try to reconnect every 1s up to 60 seconds until success
     ## Die if you can't after that
-    my $redis = Redis->new(reconnect => 60);
+    my $redis = Redis->new(reconnect => 60, every => 1000);
 
-    ## Try each 100ms upto 2 seconds (every is in milisecs)
-    my $redis = Redis->new(reconnect => 2, every => 100);
+    ## Try each 100ms upto 2 seconds (every is in nanosecond)
+    my $redis = Redis->new(reconnect => 2, every => 100_000);
 
     ## Enable connection timeout (in seconds)
     my $redis = Redis->new(cnx_timeout => 60);
@@ -1132,7 +1138,7 @@ tcp:127.0.0.1:11011
 
 The C<< reconnect >> option enables auto-reconnection mode. If we cannot
 connect to the Redis server, or if a network write fails, we enter retry mode.
-We will try a new connection every C<< every >> milliseconds (1000ms by
+We will try a new connection every C<< every >> nanoseconds (1 ms by
 default), up-to C<< reconnect >> seconds.
 
 Be aware that read errors will always thrown an exception, and will not trigger
@@ -1773,7 +1779,7 @@ The C<slowlog> command gives access to the server's slow log.
 
 =head1 ACKNOWLEDGEMENTS
 
-The following persons contributed to this project (alphabetical order):
+The following persons contributed to this project (random order):
 
 =over
 
